import {
  Body,
  Controller,
  Delete,
  Get,
  Post,
  Query,
  Request,
  UseGuards,
} from '@nestjs/common';
import { AuthGuard } from '@nestjs/passport';
import {
  ApiBearerAuth,
  ApiBody,
  ApiOperation,
  ApiQuery,
  ApiTags,
} from '@nestjs/swagger';
import { Serealize } from '../interceptors/serialize.interceptor';
import { CarreraProgramaService } from './carrera-programa.service';
import { ValidarUsuarioService } from '../validar-usuario/validar-usuario.service';
import { Operador } from '../operador/entity/operador.entity';
import { Usuario } from '../usuario/entity/usuario.entity';
import { IdInstitucionDto } from '../dto/input/id-institucion.dto';
import { CreateCarreraProgramaDto } from './dto/input/create.dto';
import { DeleteCarreraProgramaDto } from './dto/input/delete.dto';
import { MessageOutputDto } from '../dto/output/message.dto';
import { CarreraProgramaOutputDto } from './dto/output/carrera-programa.dto';

@Controller('carrera-programa')
@ApiTags('carrera-programa')
export class CarreraProgramaController {
  constructor(
    private carreraProgramaService: CarreraProgramaService,
    private validarUsuarioService: ValidarUsuarioService,
  ) {}

  @Serealize(MessageOutputDto)
  @Post()
  @UseGuards(AuthGuard('jwt'))
  @ApiOperation({
    description:
      'Endpoint que crea una asociación entre una carrera y un programa.',
  })
  @ApiBearerAuth('jwt')
  @ApiBody({
    description: 'Ambas variables son obligatorios.',
    examples: {
      ejemplo: { value: { id_institucion_carrera: 36, id_programa: 1 } },
    },
  })
  create(@Request() req, @Body() body: CreateCarreraProgramaDto) {
    const admin: Operador = req.user.operador;

    this.validarUsuarioService.validarAdmin(admin);
    return this.carreraProgramaService.create(
      admin,
      body.id_institucion_carrera,
      body.id_programa,
    );
  }

  @Serealize(MessageOutputDto)
  @Delete()
  @UseGuards(AuthGuard('jwt'))
  @ApiOperation({
    description:
      'Endpoint que borra la asociación entre una carrera y un programa.',
  })
  @ApiBearerAuth('jwt')
  @ApiBody({
    description: 'Es obligatorio mandar la variable id_carrera_programa.',
    examples: { ejemplo: { value: { id_carrera_programa: 1 } } },
  })
  delete(@Request() req, @Body() body: DeleteCarreraProgramaDto) {
    const admin: Operador = req.user.operador;

    this.validarUsuarioService.validarAdmin(admin);
    return this.carreraProgramaService.delete(admin, body.id_carrera_programa);
  }

  @Serealize(CarreraProgramaOutputDto)
  @Get()
  @UseGuards(AuthGuard('jwt'))
  @ApiOperation({
    description:
      'Endpoint que retorna todos los programas reservados para una carrera de una institución.',
  })
  @ApiBearerAuth('jwt')
  @ApiQuery({
    description: 'Id de la institución.',
    name: 'id_institucion',
    type: 'string',
  })
  get(@Request() req, @Query() query: IdInstitucionDto) {
    const usuarioOperador: Operador | Usuario =
      req.user.operador || req.user.usuario;

    if (usuarioOperador instanceof Usuario)
      this.validarUsuarioService.validarUsuario(usuarioOperador);
    else this.validarUsuarioService.validarAdmin(usuarioOperador);
    return this.carreraProgramaService.findByIdInstitucion(
      usuarioOperador,
      parseInt(query.id_institucion),
    );
  }
}
