import {
  ConflictException,
  Injectable,
  NotFoundException,
} from '@nestjs/common';
import { InjectRepository } from '@nestjs/typeorm';
import { Repository } from 'typeorm';
import { InstitucionTipoEntrada } from './entity/institucion-tipo-entrada.entity';
import { Operador } from '../operador/entity/operador.entity';
import { TipoEntrada } from './entity/tipo-entrada.entity';
import { InstitucionService } from '../institucion/institucion.service';

@Injectable()
export class InstitucionTipoEntradaService {
  constructor(
    @InjectRepository(InstitucionTipoEntrada)
    private institucionTipoEntradaRepository: Repository<InstitucionTipoEntrada>,
    @InjectRepository(TipoEntrada)
    private tipoEntradaRepository: Repository<TipoEntrada>,
    private institucionService: InstitucionService,
  ) {}

  async create(tipo_entrada: string) {
    return this.tipoEntradaRepository
      .findOne({ where: { tipo_entrada } })
      .then((existeTipoEntrada) => {
        if (existeTipoEntrada)
          throw new ConflictException('Ya existe este tipo de entrada.');
        return this.tipoEntradaRepository.save(
          this.tipoEntradaRepository.create({
            tipo_entrada,
          }),
        );
      })
      .then(async (tipoEntrada) => {
        const instituciones =
          await this.institucionService.findInfoInstitucionAll();

        for (let i = 0; i < instituciones.length; i++)
          await this.institucionTipoEntradaRepository.save(
            this.institucionTipoEntradaRepository.create({
              institucion: instituciones[i],
              tipoEntrada,
            }),
          );
        return {
          message: 'Se creó correctamente un nuevo tipo de entrada.',
        };
      });
  }

  findAll() {
    return this.tipoEntradaRepository.find({ order: { tipo_entrada: 'ASC' } });
  }

  async findAllByIdInstitucion(id_institucion: number, mostrar = false) {
    const institucion = await this.institucionService.findInfoInstitucionById(
      id_institucion,
    );
    const query = await this.institucionTipoEntradaRepository
      .createQueryBuilder('ite')
      .innerJoinAndSelect(
        'ite.institucion',
        'i',
        'i.id_institucion = :id_institucion',
        { id_institucion: institucion.id_institucion },
      )
      .innerJoinAndSelect('ite.tipoEntrada', 'te')
      .orderBy('te.tipo_entrada');

    if (mostrar) query.andWhere('ite.mostrar = 1');
    return query.getMany();
  }

  findById(id_institucion_tipo_entrada: number) {
    return this.institucionTipoEntradaRepository
      .findOne({
        join: {
          alias: 'ite',
          innerJoinAndSelect: { i: 'ite.institucion', te: 'ite.tipoEntrada' },
        },
        where: { id_institucion_tipo_entrada },
      })
      .then((institucionTipoEntrada) => {
        if (!institucionTipoEntrada)
          throw new NotFoundException(
            'No existe este id institución tipo entrada.',
          );
        return institucionTipoEntrada;
      });
  }

  findTipoEntradaById(id_tipo_entrada: number) {
    return this.tipoEntradaRepository
      .findOne({ where: { id_tipo_entrada } })
      .then((tipoEntrada) => {
        if (!tipoEntrada)
          throw new NotFoundException('No existe este id tipo de entrada.');
        return tipoEntrada;
      });
  }

  findTipoEntradaByTipoEntrada(tipo_entrada: string, validarNoExiste = true) {
    return this.tipoEntradaRepository
      .findOne({ where: { tipo_entrada } })
      .then((tipoEntrada) => {
        if (validarNoExiste && !tipoEntrada)
          throw new NotFoundException('No existe este tipo de entrada.');
        return tipoEntrada;
      });
  }

  update(admin: Operador, attrs: Partial<InstitucionTipoEntrada>) {
    return this.findById(attrs.id_institucion_tipo_entrada)
      .then((institucionTipoEntrada) => {
        if (
          admin.institucion.id_institucion !=
          institucionTipoEntrada.institucion.id_institucion
        )
          throw new ConflictException(
            'No puedes actualizar la información de este tipo de carrito porque no le pertenece a tu institución.',
          );
        Object.assign(institucionTipoEntrada, attrs);
        return this.institucionTipoEntradaRepository.save(
          institucionTipoEntrada,
        );
      })
      .then((_) => ({
        message: 'Se guardaron los cambios correctamente.',
      }));
  }
}
